@@ -31,6 +31,15 @@ class OverwriteError(Exception):
     pass
 
 
+class ExpectedValueError(Exception):
+    """Conditional write failure.
+    Raised when the expected_values argument to :meth:`DynamoDBModel.save`
+    doesn't match what is stored in the database (i.e. when somebody changed
+    the DB's version of your object behind your back).
+    """
+    pass
+
+
 class autoincrement_int(int):
     """Dummy int subclass for use in your schemas.
 
@@ -38,8 +47,9 @@ class autoincrement_int(int):
     table, new objects in your table will have an auto-incrementing primary
     key.
 
-    Note, however, that attempts to insert items with explicit, user-set values
-    for the hash key will fail.
+    Note that you can still insert items with explicit values for your primary
+    key -- the autoincrementing scheme is only used for objects with unset
+    hash_keys (or to be more precise, left set to the default value of 0).
 
     Auto-incrementing int keys are implemented by storing a special "magic"
     item in the table with the following properties:
@@ -84,6 +94,9 @@ class json_dict(dict):
             super(json_dict, self).__init__()
 
 
+_JSON_TYPES = frozenset([json_list, json_dict])
+
+
 class ConnectionBorg(object):
     """Borg that handles access to DynamoDB.
 
@@ -369,9 +382,9 @@ class DynamoDBModel(object):
         out = {}
         for (name, type_) in type(self).__schema__.iteritems():
             value = getattr(self, name)
-            if type_ in [json_list, json_dict]:
+            if type_ in _JSON_TYPES:
                 # json serialization hooks for json_* data types.
-                out[name] = json.dumps(value)
+                out[name] = json.dumps(value, sort_keys=True)
             elif value or value == 0:
                 # Yes, that part is horrible. DynamoDB can't store empty
                 # sets/strings, so we're representing them as missing
@@ -412,6 +425,11 @@ class DynamoDBModel(object):
                 # Conditional write: we're overwriting iff the value
                 # hasn't changed.
                 max_hash_item.put({"__max_hash_key__": max_hash_key})
+                # We just reserved that value for the hash key
+                item[item.hash_key_name] = max_hash_key + 1
+                # If despite our checks someone manually inserted an item at
+                # that primary key, we'll retry with the next available value.
+                item.put({item.hash_key_name: False})
                 break
             except DynamoDBResponseError as e:
                 if e.error_code != "ConditionalCheckFailedException":
@@ -419,17 +437,25 @@ class DynamoDBModel(object):
                     raise
                 # The max key has changed (concurrent write): retry.
 
-        # We just reserved that value for the hash key
-        item[item.hash_key_name] = max_hash_key + 1
-        item.put()
-
-    def save(self, allow_overwrite=True):
+    def save(self, allow_overwrite=True, expected_values=None):
         """Save the object to the database.
 
         This method may be used both to insert a new object in the DB, or to
-        update an existing one (iff allow_overwrite == True -- otherwise,
-        the operation fails with OverwriteError).
+        update an existing one (iff allow_overwrite == True).
+
+        :param allow_overwrite: If False, the method will only succeed if this
+            object's primary keys don't exist in the database (otherwise,
+            :class:`OverwriteError` is raised).
+
+        :param expected_values: dict of expected attribute values -- if any
+            one of these values doesn't match what is in the database
+            (i.e. someone went ahead and modified the object in the DB behind
+            your back), the operation fails and raises
+            :class:`ExpectedValueError`.
         """
+        if expected_values is None:
+            expected_values = {}
+
         cls = type(self)
         schema = cls.__schema__
         hash_key = cls.__hash_key__
@@ -445,17 +471,29 @@ class DynamoDBModel(object):
             setattr(self, hash_key, item[hash_key])
         else:
             if allow_overwrite:
-                expected_values = None
+                expected_db_values = {}
+                schema = type(self).__schema__
+                for (name, value) in expected_values.iteritems():
+                    type_ = schema[name]
+                    # This bit can probably be factored out with to_db_dict
+                    if type_ in _JSON_TYPES:
+                        expected_db_values[name] = json.dumps(value, sort_keys=True)
+                    elif value or value == 0:
+                        expected_db_values[name] = value
             else:
                 # Forbid overwrites: do a conditional write on
                 # "this hash_key doesn't exist"
-                expected_values = {hash_key: False}
+                expected_db_values = {hash_key: False}
                 if range_key:
-                    expected_values[range_key] = False
+                    expected_db_values[range_key] = False
             try:
-                item.put(expected_values)
+                item.put(expected_db_values)
             except DynamoDBResponseError as e:
                 if e.error_code == "ConditionalCheckFailedException":
+                    if allow_overwrite:
+                        # Failed conditional write
+                        raise ExpectedValueError(item)
+                    # Forbidden overwrite
                     raise OverwriteError(item)
                 # Unhandled exception
                 raise
